from . import attrs, codegen, prelude
from ._filters import (
    AuxData,
    combine,
    combine_leaves,
    is_data,
    is_leaf,
    partition,
    partition_leaves,
)
from .attrs import array, auto, define, field, frozen, register_fieldz, static
from .codegen import codegen_pytree_functions, register_generic
from .prelude import PyTreeProxy, register_pytree_prelude

__all__ = [
    "AuxData",
    "PyTreeProxy",
    "array",
    "attrs",
    "auto",
    "codegen",
    "codegen_pytree_functions",
    "combine",
    "combine_leaves",
    "define",
    "field",
    "frozen",
    "is_data",
    "is_leaf",
    "partition",
    "partition_leaves",
    "prelude",
    "register_fieldz",
    "register_generic",
    "register_pytree_prelude",
    "static",
]



from collections.abc import Iterable
from typing import Any

import jax
import jax.tree_util as jtu
from jax import Array
from jax._src.tree_util import _registry
from liblaf import grapes
from typing_extensions import TypeIs


@jtu.register_static
@grapes.attrs.frozen
class AuxData[T]:
    meta_leaves: tuple[Any, ...]
    treedef: Any


def is_data(obj: Any) -> bool:
    return obj is None or isinstance(obj, Array) or type(obj) in _registry


def is_leaf(obj: Any) -> TypeIs[Array | None]:
    return obj is None or isinstance(obj, Array)


def combine[T](data_leaves: Iterable[Array | None], aux: AuxData[T]) -> T:
    leaves: list[Any] = combine_leaves(data_leaves, aux.meta_leaves)
    return jax.tree.unflatten(aux.treedef, leaves)


def combine_leaves(
    data_leaves: Iterable[Array | None], meta_leaves: Iterable[Any]
) -> list[Any]:
    return [
        data_leaf if meta_leaf is None else meta_leaf
        for data_leaf, meta_leaf in zip(data_leaves, meta_leaves, strict=True)
    ]


def partition(obj: Any) -> tuple[list[Array | None], AuxData]:
    leaves: list[Any]
    treedef: Any
    leaves, treedef = jax.tree.flatten(obj)
    data_leaves: list[Array | None]
    meta_leaves: list[Any]
    data_leaves, meta_leaves = partition_leaves(leaves)
    return data_leaves, AuxData(tuple(meta_leaves), treedef)


def partition_leaves(leaves: list[Any]) -> tuple[list[Array | None], list[Any]]:
    data_leaves: list[Array | None] = []
    meta_leaves: list[Any] = []
    for leaf in leaves:
        if is_leaf(leaf):
            data_leaves.append(leaf)
            meta_leaves.append(None)
        else:
            data_leaves.append(None)
            meta_leaves.append(leaf)
    return data_leaves, meta_leaves



import ast
import linecache
import types
from collections.abc import Callable, Sequence
from typing import Any, NamedTuple

import jax.tree_util as jtu

from jarp.tree._filters import is_data

from ._codegen import codegen_flatten, codegen_flatten_with_keys, codegen_unflatten


class PyTreeFunctions(NamedTuple):
    flatten: Callable
    unflatten: Callable
    flatten_with_keys: Callable


def codegen_pytree_functions(
    cls: type,
    data_fields: Sequence[str] = (),
    meta_fields: Sequence[str] = (),
    auto_fields: Sequence[str] = (),
    *,
    filter_spec: Callable[[Any], bool] = is_data,
    bypass_setattr: bool | None = None,
) -> PyTreeFunctions:
    if bypass_setattr is None:
        bypass_setattr = cls.__setattr__ is not object.__setattr__
    flatten_def: ast.FunctionDef = codegen_flatten(
        data_fields, meta_fields, auto_fields
    )
    flatten_with_keys_def: ast.FunctionDef = codegen_flatten_with_keys(
        data_fields, meta_fields, auto_fields
    )
    unflatten_def: ast.FunctionDef = codegen_unflatten(
        data_fields, meta_fields, auto_fields, bypass_setattr=bypass_setattr
    )
    module: ast.Module = ast.Module(
        body=[flatten_def, flatten_with_keys_def, unflatten_def], type_ignores=[]
    )
    module = ast.fix_missing_locations(module)
    source: str = ast.unparse(module)
    namespace: dict = {
        "_cls": cls,
        "_filter_spec": filter_spec,
        "_object_new": object.__new__,
        "_object_setattr": object.__setattr__,
        **_make_keys((*data_fields, *meta_fields, *auto_fields)),
    }
    filename: str = _make_filename(cls)
    # use unparse source so we have correct source code locations
    code: types.CodeType = compile(source, filename, "exec")
    exec(code, namespace)  # noqa: S102
    _update_linecache(source, filename)
    return PyTreeFunctions(
        _add_dunder(cls, namespace["flatten"]),
        _add_dunder(cls, namespace["unflatten"]),
        _add_dunder(cls, namespace["flatten_with_keys"]),
    )


def register_generic(
    cls: type,
    data_fields: Sequence[str] = (),
    meta_fields: Sequence[str] = (),
    auto_fields: Sequence[str] = (),
    *,
    filter_spec: Callable[[Any], bool] = is_data,
    bypass_setattr: bool | None = None,
) -> None:
    flatten: Callable
    unflatten: Callable
    flatten_with_keys: Callable
    flatten, unflatten, flatten_with_keys = codegen_pytree_functions(
        cls,
        data_fields,
        meta_fields,
        auto_fields,
        filter_spec=filter_spec,
        bypass_setattr=bypass_setattr,
    )
    jtu.register_pytree_node(cls, flatten, unflatten, flatten_with_keys)


def _add_dunder[C: Callable](cls: type, func: C) -> C:
    func.__module__ = cls.__module__
    return func


def _make_filename(cls: type) -> str:
    return f"<jarp generated functions {cls.__module__}.{cls.__qualname__}>"


def _make_keys(fields: Sequence[str]) -> dict[str, Any]:
    return {f"_{name}_key": jtu.GetAttrKey(name) for name in fields}


def _update_linecache(source: str, filename: str) -> None:
    linecache.cache[filename] = (
        len(source),  # size
        None,  # mtime
        source.splitlines(keepends=True),  # lines
        filename,  # fullname
    )



import types
from collections.abc import Callable
from typing import Any

import jax.tree_util as jtu


def _method_flatten(obj: types.MethodType) -> tuple[tuple[object], Callable[..., Any]]:
    return (obj.__self__,), obj.__func__


def _method_flatten_with_keys(
    obj: types.MethodType,
) -> tuple[tuple[tuple[Any, object]], Callable[..., Any]]:
    return ((jtu.GetAttrKey("__self__"), obj.__self__),), obj.__func__


def _method_unflatten(
    aux: Callable[..., Any], children: tuple[object]
) -> types.MethodType:
    return types.MethodType(aux, children[0])


def register_pytree_method() -> None:
    jtu.register_pytree_node(
        types.MethodType, _method_flatten, _method_unflatten, _method_flatten_with_keys
    )



import re
import sys

import jax
import warp as wp


def _default_float_type() -> type:
    return wp.float64 if jax.config.read("jax_enable_x64") else wp.float32


def vector(length: int) -> type:
    return wp.types.vector(length, _default_float_type())


def matrix(shape: tuple[int, int]) -> type:
    return wp.types.matrix(shape, _default_float_type())


def __getattr__(name: str) -> type:
    if name == "float_":
        return _default_float_type()
    result: re.Match[str] | None = re.fullmatch(r"vec(?P<length>[1-9])", name)
    if result is not None:
        length = int(result.group("length"))
        return wp.types.vector(length, _default_float_type())
    result: re.Match[str] | None = re.fullmatch(
        r"mat(?P<rows>[1-9])(?P<cols>[1-9])", name
    )
    if result is not None:
        rows = int(result.group("rows"))
        cols = int(result.group("cols"))
        return wp.types.matrix((rows, cols), _default_float_type())
    msg: str = f"module '{__name__}' has no attribute '{name}'"
    raise AttributeError(msg, name=name, obj=sys.modules[__name__])



from . import toolz, tree, warp
from ._jit import jit
from ._version import __version__, __version_tuple__
from .tree import (
    PyTreeProxy,
    array,
    auto,
    define,
    field,
    frozen,
    register_pytree_prelude,
    static,
)
from .warp import jax_callable, jax_kernel, to_warp

__all__ = [
    "PyTreeProxy",
    "__version__",
    "__version_tuple__",
    "array",
    "auto",
    "define",
    "field",
    "frozen",
    "jax_callable",
    "jax_kernel",
    "jit",
    "register_pytree_prelude",
    "static",
    "to_warp",
    "toolz",
    "tree",
    "warp",
]



import functools
from collections.abc import Callable, Iterable, Sequence
from typing import Any, Protocol, TypedDict, Unpack, cast, overload

import jax.tree_util as jtu
import warp as wp
import warp.jax_experimental
from jax import Array
from warp._src.jax_experimental.ffi import FfiKernel as _WarpFfiKernel
from warp._src.jax_experimental.ffi import ModulePreloadMode

from jarp import tree

from ._types import ArgTypes, ShapeLike, VmapMethod, WarpScalarDType
from ._utils import dtypes_from_args


class JaxKernelOptions(TypedDict, total=False):
    num_outputs: int
    vmap_method: VmapMethod
    launch_dims: ShapeLike | None
    output_dims: ShapeLike | dict[str, ShapeLike] | None  # Mapping won't work with Warp
    in_out_argnames: Iterable[str]
    module_preload_mode: ModulePreloadMode
    enable_backward: bool


class JaxKernelCallOptions(TypedDict, total=False):
    output_dims: ShapeLike | dict[str, ShapeLike] | None  # Mapping won't work with Warp
    launch_dims: ShapeLike | None
    vmap_method: VmapMethod | None


class FfiKernelProtocol(Protocol):
    def __call__(
        self, *args: Array, **kwargs: Unpack[JaxKernelCallOptions]
    ) -> Sequence[Array]: ...


@tree.frozen(static=True)
class _FfiKernel(FfiKernelProtocol):
    kernel: wp.Kernel
    options: JaxKernelOptions
    arg_types_factory: Callable[..., ArgTypes]

    def __call__(
        self, *args: Array, **kwargs: Unpack[JaxKernelCallOptions]
    ) -> Sequence[Array]:
        warp_dtypes: list[WarpScalarDType] = dtypes_from_args(*args)
        arg_types: ArgTypes = self.arg_types_factory(*warp_dtypes)
        kernel: wp.Kernel = wp.overload(self.kernel, arg_types)
        jax_kernel: Callable[..., Sequence[Array]] = warp.jax_experimental.jax_kernel(
            kernel, **self.options
        )
        return jax_kernel(*args, **kwargs)


# we annotate `kernel` as `Callable` instead of `wp.Kernel` because `@wp.kernel`
# does not have good typing
@overload
def jax_kernel(
    *,
    arg_types_factory: ArgTypes | Callable[[WarpScalarDType], ArgTypes] | None = None,
    **kwargs: Unpack[JaxKernelOptions],
) -> Callable[[Callable], FfiKernelProtocol]: ...
@overload
def jax_kernel(
    kernel: Callable,
    *,
    arg_types_factory: ArgTypes | Callable[[WarpScalarDType], ArgTypes] | None = None,
    **kwargs: Unpack[JaxKernelOptions],
) -> FfiKernelProtocol: ...
def jax_kernel(
    kernel: Callable | None = None,
    *,
    arg_types_factory: ArgTypes | Callable[[WarpScalarDType], ArgTypes] | None = None,
    **kwargs: Unpack[JaxKernelOptions],
) -> Any:
    if kernel is None:
        return functools.partial(
            jax_kernel, arg_types_factory=arg_types_factory, **kwargs
        )
    if arg_types_factory is None:
        return warp.jax_experimental.jax_kernel(kernel, **kwargs)
    if not callable(arg_types_factory):
        return warp.jax_experimental.jax_kernel(
            wp.overload(kernel, arg_types_factory), **kwargs
        )
    return _FfiKernel(
        kernel=cast("wp.Kernel", kernel),
        options=kwargs,
        arg_types_factory=arg_types_factory,
    )


jtu.register_static(_WarpFfiKernel)



from ast import (
    Assign,
    Attribute,
    Call,
    Compare,
    Constant,
    FunctionDef,
    IfExp,
    Is,
    Load,
    Name,
    Return,
    Store,
    Tuple,
    arg,
    expr,
    stmt,
)
from collections.abc import Sequence

from ._utils import (
    codegen_aux,
    codegen_children,
    codegen_function_def,
    codegen_object_setattr,
    codegen_partition,
    codegen_unpack,
)


def codegen_flatten(
    data_fields: Sequence[str], meta_fields: Sequence[str], auto_fields: Sequence[str]
) -> FunctionDef:
    body: list[stmt] = codegen_partition(auto_fields)
    children: list[expr] = codegen_children(data_fields, auto_fields)
    aux: list[expr] = codegen_aux(meta_fields, auto_fields)
    body.append(Return(Tuple([Tuple(children, Load()), Tuple(aux, Load())], Load())))
    return codegen_function_def("flatten", [arg("obj")], body)


def codegen_flatten_with_keys(
    data_fields: Sequence[str], meta_fields: Sequence[str], auto_fields: Sequence[str]
) -> FunctionDef:
    body: list[stmt] = codegen_partition(auto_fields)
    children: list[expr] = codegen_children(data_fields, auto_fields)
    aux: list[expr] = codegen_aux(meta_fields, auto_fields)
    keys: list[expr] = [
        Name(f"_{name}_key", Load()) for name in (*data_fields, *auto_fields)
    ]
    children_with_keys: list[expr] = [
        Tuple([key, child], Load()) for key, child in zip(keys, children, strict=True)
    ]
    body.append(
        Return(Tuple([Tuple(children_with_keys, Load()), Tuple(aux, Load())], Load()))
    )
    return codegen_function_def("flatten_with_keys", [arg("obj")], body)


def codegen_unflatten(
    data_fields: Sequence[str],
    meta_fields: Sequence[str],
    auto_fields: Sequence[str],
    *,
    bypass_setattr: bool = False,
) -> FunctionDef:
    body: list[stmt] = [
        Assign(
            [Name("obj", Store())],
            Call(Name("_object_new", Load()), [Name("_cls", Load())], []),
        )
    ]
    if bypass_setattr:
        body.extend(_codegen_unflatten_bypass(data_fields, meta_fields, auto_fields))
    else:
        body.extend(_codegen_unflatten_direct(data_fields, meta_fields, auto_fields))
    body.append(Return(Name("obj", Load())))
    return codegen_function_def("unflatten", [arg("aux"), arg("children")], body)


def _codegen_unflatten_bypass(
    data_fields: Sequence[str], meta_fields: Sequence[str], auto_fields: Sequence[str]
) -> list[stmt]:
    body: list[stmt] = [
        # _c_meta, _d_meta, ..., _e_meta, _f_meta = aux
        codegen_unpack(
            (),
            [f"_{name}_meta" for name in (*meta_fields, *auto_fields)],
            Name("aux", Load()),
        ),
        # _a_data, _b_data, ..., _e_data, _f_data = children
        codegen_unpack(
            (),
            [f"_{name}_data" for name in (*data_fields, *auto_fields)],
            Name("children", Load()),
        ),
    ]
    body.extend(
        # _object_setattr(obj, "name", _name_data)
        codegen_object_setattr(name, Name(f"_{name}_data", Load()))
        for name in data_fields
    )
    body.extend(
        # _object_setattr(obj, "name", _name_meta)
        codegen_object_setattr(name, Name(f"_{name}_meta", Load()))
        for name in meta_fields
    )
    body.extend(
        # _object_setattr(obj, "name", _name_data if _name_meta is None else _name_meta)
        codegen_object_setattr(
            name,
            # _name_data if _name_meta is None else _name_meta
            IfExp(
                # _name_meta is None
                Compare(Name(f"_{name}_meta", Load()), [Is()], [Constant(None)]),
                Name(f"_{name}_data", Load()),
                Name(f"_{name}_meta", Load()),
            ),
        )
        for name in auto_fields
    )
    return body


def _codegen_unflatten_direct(
    data_fields: Sequence[str], meta_fields: Sequence[str], auto_fields: Sequence[str]
) -> list[stmt]:
    body: list[stmt] = [
        # obj.c, obj.d, ..., _e_meta, _f_meta, ... = aux
        codegen_unpack(
            meta_fields, [f"_{name}_meta" for name in auto_fields], Name("aux", Load())
        ),
        # obj.a, obj.b, ..., _e_data, _f_data, ... = children
        codegen_unpack(
            data_fields,
            [f"_{name}_data" for name in auto_fields],
            Name("children", Load()),
        ),
    ]
    body.extend(
        # obj.name = _name_data if _name_meta is None else _name_meta
        Assign(
            [Attribute(Name("obj", Load()), name, Store())],
            # _name_data if _name_meta is None else _name_meta
            IfExp(
                # _name_meta is None
                Compare(Name(f"_{name}_meta", Load()), [Is()], [Constant(None)]),
                Name(f"_{name}_data", Load()),
                Name(f"_{name}_meta", Load()),
            ),
        )
        for name in auto_fields
    )
    return body



from typing import Any

import jax
import numpy as np
import warp as wp

def to_warp(
    arr: wp.array | np.ndarray | jax.Array,
    dtype: tuple[int, Any] | tuple[int, int, Any] | Any = ...,
    *,
    requires_grad: bool = ...,
    **kwargs,
) -> wp.array: ...



from ast import (
    Assign,
    Attribute,
    Call,
    Constant,
    Expr,
    FunctionDef,
    IfExp,
    Load,
    Name,
    Store,
    Tuple,
    arg,
    arguments,
    expr,
    stmt,
)
from collections.abc import Sequence


def codegen_aux(meta_fields: Sequence[str], auto_fields: Sequence[str]) -> list[expr]:
    return _codegen_children(meta_fields, auto_fields, suffix="_meta")


def codegen_children(
    data_fields: Sequence[str], auto_fields: Sequence[str]
) -> list[expr]:
    return _codegen_children(data_fields, auto_fields, suffix="_data")


def codegen_function_def(name: str, args: list[arg], body: list[stmt]) -> FunctionDef:
    # def name(args...): body...
    return FunctionDef(
        name=name,
        args=arguments(
            posonlyargs=[],
            args=args,
            vararg=None,
            kwonlyargs=[],
            kw_defaults=[],
            kwarg=None,
            defaults=[],
        ),
        body=body,
        decorator_list=[],
        returns=None,
        type_comment=None,
        type_params=[],
    )


def codegen_object_setattr(name: str, value: expr) -> Expr:
    # _object_setattr(obj, "name", value)
    return Expr(
        Call(
            Name("_object_setattr", Load()),
            [Name("obj", Load()), Constant(name), value],
            [],
        )
    )


def codegen_partition(auto_fields: Sequence[str]) -> list[stmt]:
    """.

    ```python
    _a_auto = obj.a
    _b_auto = obj.b
    ...
    _a_data, _a_meta = (_a_auto, None) if _filter_spec(_a_auto) else (None, _a_auto)
    _b_data, _b_meta = (_b_auto, None) if _filter_spec(_b_auto) else (None, _b_auto)
    ...
    ```
    """
    body: list[stmt] = [
        # _name_auto = obj.name
        Assign(
            [Name(f"_{name}_auto", Store())],
            Attribute(Name("obj", Load()), name, Load()),
        )
        for name in auto_fields
    ]
    body.extend(
        # _name_data, _name_meta = (_name_auto, None) if _filter_spec(_name_auto) else (None, _name_auto)
        Assign(
            [
                Tuple(
                    [Name(f"_{name}_data", Store()), Name(f"_{name}_meta", Store())],
                    Store(),
                )
            ],
            IfExp(
                Call(Name("_filter_spec", Load()), [Name(f"_{name}_auto", Load())], []),
                Tuple([Name(f"_{name}_auto", Load()), Constant(None)], Load()),
                Tuple([Constant(None), Name(f"_{name}_auto", Load())], Load()),
            ),
        )
        for name in auto_fields
    )
    return body


def codegen_unpack(attrs: Sequence[str], names: Sequence[str], value: expr) -> Assign:
    # (obj.attr..., name...) = value
    targets: list[expr] = [
        Attribute(Name("obj", Load()), attr, Store()) for attr in attrs
    ]
    targets.extend(Name(name, Store()) for name in names)
    return Assign([Tuple(targets, Store())], value)


def _codegen_children(
    fields: Sequence[str], auto_fields: Sequence[str], suffix: str
) -> list[expr]:
    children: list[expr] = []
    children.extend(Attribute(Name("obj", Load()), name, Load()) for name in fields)
    children.extend(Name(f"_{name}{suffix}", Load()) for name in auto_fields)
    return children



import jax
import warp as wp
from jax import Array

from ._types import WarpScalarDType


def dtypes_from_args(*args: Array) -> list[WarpScalarDType]:
    return [wp.dtype_from_jax(jax.dtypes.result_type(arr)) for arr in args]



from collections.abc import Sequence
from typing import Any, Literal

type ArgTypes = dict[str, WarpDType] | list[WarpDType] | None
type ShapeLike = int | Sequence[int]
type VmapMethod = Literal["broadcast_all", "expand_dims", "sequential"]
type WarpDType = Any
type WarpScalarDType = Any



from collections.abc import Iterable

import jax.tree_util as jtu
import wrapt
from jax import Array

from jarp.tree._filters import AuxData, combine, partition


class PyTreeProxy[T](wrapt.BaseObjectProxy):
    __wrapped__: T


def _proxy_flatten[T](obj: PyTreeProxy[T]) -> tuple[list[Array | None], AuxData[T]]:
    return partition(obj.__wrapped__)


def _proxy_unflatten[T](
    aux: AuxData[T], children: Iterable[Array | None]
) -> PyTreeProxy[T]:
    return PyTreeProxy(combine(children, aux))


jtu.register_pytree_node(PyTreeProxy, _proxy_flatten, _proxy_unflatten)



import functools
from collections.abc import Sequence
from typing import Any

import jax
import numpy as np
import warp as wp


@functools.singledispatch
def to_warp(arr: Any, *_args, **_kwargs) -> wp.array:
    raise TypeError(arr)


def _convert_dtype(dtype: Any, arr_shape: Sequence[int], arr_dtype: Any) -> Any:
    match dtype:
        case (length, dtype):
            if length == -1:
                length: int = arr_shape[-1]
            if dtype is None or dtype is Any:
                dtype = arr_dtype
            return wp.types.vector(length, dtype)
        case (rows, cols, dtype):
            if rows == -1:
                rows: int = arr_shape[-2]
            if cols == -1:
                cols: int = arr_shape[-1]
            if dtype is None or dtype is Any:
                dtype = arr_dtype
            return wp.types.matrix((rows, cols), dtype)
        case _:
            return dtype


@to_warp.register(np.ndarray)
def _to_warp_numpy(arr: np.ndarray, dtype: Any = None, **kwargs) -> wp.array:
    dtype = _convert_dtype(dtype, arr.shape, wp.dtype_from_numpy(arr.dtype))
    return wp.from_numpy(arr, dtype, **kwargs)


@to_warp.register(jax.Array)
def _to_warp_jax(arr: jax.Array, dtype: Any = None, **kwargs) -> wp.array:
    dtype = _convert_dtype(dtype, arr.shape, wp.dtype_from_jax(arr.dtype))
    requires_grad: bool = kwargs.pop("requires_grad", False)
    arr_wp: wp.array = wp.from_jax(arr, dtype, **kwargs)
    if requires_grad:
        arr_wp.requires_grad = True
    return arr_wp



from . import types
from ._jax_callable import (
    FfiCallableProtocol,
    JaxCallableCallOptions,
    JaxCallableOptions,
    jax_callable,
)
from ._jax_kernel import (
    FfiKernelProtocol,
    JaxKernelCallOptions,
    JaxKernelOptions,
    jax_kernel,
)
from ._to_warp import to_warp

__all__ = [
    "FfiCallableProtocol",
    "FfiKernelProtocol",
    "JaxCallableCallOptions",
    "JaxCallableOptions",
    "JaxKernelCallOptions",
    "JaxKernelOptions",
    "jax_callable",
    "jax_kernel",
    "to_warp",
    "types",
]



from collections.abc import Callable, Iterable
from typing import Any, Protocol

import jax.tree_util as jtu
import tlz
from liblaf import grapes


@grapes.wraps(tlz.compose_left)
def compose_left(*args, **kwargs) -> Any:
    return tlz.compose_left(*args, **kwargs)


class Compose(Protocol):
    @property
    def first(self) -> Callable[..., Any]: ...
    @property
    def funcs(self) -> Iterable[Callable[..., Any]]: ...


def _compose_flatten(obj: Compose) -> tuple[list[Callable[..., Any]], None]:
    return [obj.first, *obj.funcs], None


def _compose_unflatten(
    _aux: None, children: list[Callable[..., Any]]
) -> Callable[..., Any]:
    return tlz.compose(*children)


jtu.register_pytree_node(
    tlz.functoolz.Compose,  # pyright: ignore[reportAttributeAccessIssue]
    _compose_flatten,
    _compose_unflatten,  # pyright: ignore[reportArgumentType]
)



from ._define import define, frozen
from ._field_specifiers import array, auto, field, static
from ._register import register_fieldz

__all__ = ["array", "auto", "define", "field", "frozen", "register_fieldz", "static"]



from collections.abc import Callable, Sequence
from typing import Any

import fieldz

from jarp.tree._filters import is_data
from jarp.tree.codegen import register_generic


def register_fieldz[T: type](
    cls: T,
    data_fields: Sequence[str] | None = None,
    meta_fields: Sequence[str] | None = None,
    auto_fields: Sequence[str] | None = None,
    *,
    filter_spec: Callable[[Any], bool] = is_data,
    bypass_setattr: bool | None = None,
) -> T:
    if data_fields is None:
        data_fields = _filter_field_names(_filter_data, cls)
    if meta_fields is None:
        meta_fields = _filter_field_names(_filter_meta, cls)
    if auto_fields is None:
        auto_fields = _filter_field_names(_filter_auto, cls)
    register_generic(
        cls,
        data_fields,
        meta_fields,
        auto_fields,
        filter_spec=filter_spec,
        bypass_setattr=bypass_setattr,
    )
    return cls


def _filter_field_names(
    function: Callable[[fieldz.Field], bool], cls: type
) -> list[str]:
    return [f.name for f in fieldz.fields(cls) if function(f)]


def _filter_data(field: fieldz.Field) -> bool:
    return not (_filter_auto(field) or _filter_meta(field))


def _filter_meta(field: fieldz.Field) -> bool:
    return field.metadata.get("static", False)


def _filter_auto(field: fieldz.Field) -> bool:
    return field.metadata.get("auto", False)



import functools

from ._method import register_pytree_method
from ._warp import register_warp_array


@functools.cache  # run only once
def register_pytree_prelude() -> None:
    register_pytree_method()
    register_warp_array()



from ._codegen import codegen_flatten, codegen_flatten_with_keys, codegen_unflatten
from ._compile import PyTreeFunctions, codegen_pytree_functions, register_generic

__all__ = [
    "PyTreeFunctions",
    "codegen_flatten",
    "codegen_flatten_with_keys",
    "codegen_pytree_functions",
    "codegen_unflatten",
    "register_generic",
]



import functools
from collections.abc import Callable, Iterable, Sequence
from typing import Any, Literal, Protocol, TypedDict, Unpack, overload

import jax.tree_util as jtu
import warp.jax_experimental
from jax import Array
from warp._src.jax_experimental.ffi import FfiCallable as _WarpFfiCallable
from warp._src.jax_experimental.ffi import ModulePreloadMode
from warp.jax_experimental import GraphMode

from jarp import tree

from ._types import ShapeLike, VmapMethod, WarpScalarDType
from ._utils import dtypes_from_args

type _FfiCallableFunction = Callable[..., None]
type _FfiCallableFactory = Callable[..., _FfiCallableFunction]


class JaxCallableOptions(TypedDict, total=False):
    num_outputs: int
    graph_mode: GraphMode
    vmap_method: str | None
    output_dims: dict[str, ShapeLike] | None  # Mapping won't work with Warp
    in_out_argnames: Iterable[str]
    stage_in_argnames: Iterable[str]
    stage_out_argnames: Iterable[str]
    graph_cache_max: int | None
    module_preload_mode: ModulePreloadMode


class JaxCallableCallOptions(TypedDict, total=False):
    output_dims: ShapeLike | dict[str, ShapeLike] | None  # Mapping won't work with Warp
    vmap_method: VmapMethod | None


class FfiCallableProtocol(Protocol):
    def __call__(
        self, *args: Array, **kwargs: Unpack[JaxCallableCallOptions]
    ) -> Sequence[Array]: ...


@tree.frozen(static=True)
class _FfiCallable(FfiCallableProtocol):
    factory: _FfiCallableFactory
    options: JaxCallableOptions

    def __call__(
        self, *args: Array, **kwargs: Unpack[JaxCallableCallOptions]
    ) -> Sequence[Array]:
        warp_dtypes: list[WarpScalarDType] = dtypes_from_args(*args)
        func: _FfiCallableFunction = self.factory(*warp_dtypes)
        jax_callable: Callable[..., Sequence[Array]] = (
            warp.jax_experimental.jax_callable(func, **self.options)
        )
        return jax_callable(*args, **kwargs)


@overload
def jax_callable(
    func: _FfiCallableFunction,
    *,
    generic: Literal[False] = False,
    **kwargs: Unpack[JaxCallableOptions],
) -> FfiCallableProtocol: ...
@overload
def jax_callable(
    *, generic: Literal[False] = False, **kwargs: Unpack[JaxCallableOptions]
) -> Callable[[_FfiCallableFunction], FfiCallableProtocol]: ...
@overload
def jax_callable(
    func: _FfiCallableFactory,
    *,
    generic: Literal[True],
    **kwargs: Unpack[JaxCallableOptions],
) -> _FfiCallable: ...
@overload
def jax_callable(
    *, generic: Literal[True], **kwargs: Unpack[JaxCallableOptions]
) -> Callable[[_FfiCallableFactory], _FfiCallable]: ...
def jax_callable(
    func: Callable | None = None,
    *,
    generic: bool = False,
    **kwargs: Unpack[JaxCallableOptions],
) -> Any:
    if func is None:
        return functools.partial(jax_callable, generic=generic, **kwargs)
    if not generic:
        return warp.jax_experimental.jax_callable(func, **kwargs)
    return _FfiCallable(factory=func, options=kwargs)


jtu.register_static(_WarpFfiCallable)



from __future__ import annotations

from collections.abc import Callable, Mapping
from typing import Any, TypedDict, Unpack

import attrs
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike
from liblaf import grapes


class FieldOptions[T](TypedDict, total=False):
    default: T
    validator: attrs._ValidatorArgType[T] | None
    repr: attrs._ReprArgType
    hash: bool | None
    init: bool
    metadata: Mapping[Any, Any] | None
    converter: (
        attrs._ConverterType
        | list[attrs._ConverterType]
        | tuple[attrs._ConverterType, ...]
        | None
    )
    factory: Callable[[], T] | None
    kw_only: bool | None
    eq: attrs._EqOrderType | None
    order: attrs._EqOrderType | None
    on_setattr: attrs._OnSetAttrArgType | None
    alias: str | None
    type: type | None


def array(**kwargs: Unpack[FieldOptions[ArrayLike | None]]) -> Array:
    if "default" in kwargs and "factory" not in kwargs:
        default: ArrayLike | None = kwargs.pop("default")  # pyright: ignore[reportAssignmentType]
        if default is not None:
            default = jnp.asarray(default)
        kwargs["factory"] = lambda: default
    return field(**kwargs)  # pyright: ignore[reportArgumentType, reportCallIssue]


@grapes.wraps(attrs.field)
def auto(**kwargs) -> Any:
    kwargs.setdefault("auto", True)
    return field(**kwargs)


@grapes.wraps(attrs.field)
def field(**kwargs) -> Any:
    auto: bool = kwargs.pop("auto", False)
    static: bool = kwargs.pop("static", False)
    assert not (auto and static), "Field cannot be both auto and static"
    if auto:
        kwargs["metadata"] = {"auto": auto, **kwargs.pop("metadata", {})}
    if static:
        kwargs["metadata"] = {"static": static, **kwargs.pop("metadata", {})}
    return attrs.field(**kwargs)


@grapes.wraps(attrs.field)
def static(**kwargs) -> Any:
    kwargs.setdefault("static", True)
    return field(**kwargs)
